/*
 *   <one line to give the program's name and a brief idea of what it does.>
 *   Copyright (C) 2011  Mohammadhossein Malmir & Amirahmad Barekatein :
 *			 <mhmalmir@gmail.com> & <amirahmad.barekatein@gmail.com>
 * 
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 * 
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "bhv_mark.h"
#include <rcsc/geom/vector_2d.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/player/soccer_action.h>
#include <vector>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/audio_memory.h>

#define MAX_THEIR_PLAYERS 11
#define MAXSIZE 25
#define INF 1000000000

bool Bhv_Mark::ball_our_possession;
int  Bhv_Mark::cycle_ball_poss;
int  Bhv_Mark::last_blocker; 
int  Bhv_Mark::last_marker_unum[7];
int  Bhv_Mark::last_say_cycle;

bool Bhv_Mark::markExecute( PlayerAgent * agent ,
			    int & opp_target_unum ,
			    bool & is_blocker )
{
 const WorldModel & wm = agent->world();
 int blocker_unum  = 0 ; 
 
 if(wm.self().goalie() )
 {
   if( wm.time().cycle() < 5 )
   {
     last_say_cycle = 0;
     for(int i = 0 ; i < 7 ; i ++ )
       last_marker_unum[i] = 0;
   }
   
   if( !(!wm.opponentsFromBall().empty() 
     && wm.opponentsFromBall().front()->ghostCount() < 7 
     && ( wm.interceptTable()->opponentReachCycle() < wm.interceptTable()->teammateReachCycle() + 2 
       || wm.existKickableOpponent() ) ) )
     return false;

   
   if( !block_decision( agent , blocker_unum ) )
     blocker_unum = 0;


   const PlayerPtrCont & offenders = agent->world().opponentsFromSelf();
   const PlayerPtrCont & markers = agent->world().teammatesFromSelf();


    if( ( offenders.empty() || markers.empty() ) && blocker_unum == 0 )
    {
      opp_target_unum = 0 ;
      return false;
    }
    
    PlayerPtrCont::const_iterator off_it_end = offenders.end();
    PlayerPtrCont::const_iterator mrk_it_end = markers.end();
    for (PlayerPtrCont::const_iterator off_it = offenders.begin();
	 off_it != off_it_end;
	 off_it ++)
    {
      
	  double  average = 0.0  ; 
	  const int score_threshold = -32000 ;

	    if( ( *off_it ) == wm.opponentsFromBall().front() )
	    {
		  continue ; 
	    }

	  for ( PlayerPtrCont::const_iterator mrk_it = markers.begin();
		mrk_it != mrk_it_end;
		mrk_it ++)
	  {
	      
	      cached_edges.push_back( MarkEdge((*mrk_it), (*off_it), agent) );
	  }
      }
      
      bool best_edge = false;
      
      
      if(!cached_edges.empty() )
      {
	best_edge = get_best_edge(agent);
      }
      
      if( blocker_unum !=0 || best_edge )
	return say(agent , blocker_unum );
  }
  
 else
 {
     return hear_n_parsing( agent , opp_target_unum , is_blocker );
 }

return false ;
}

/*!*****************************************************************************************************************/
bool Bhv_Mark::hear_n_parsing(PlayerAgent* agent , int & opp_target_unum , bool & is_blocker )
{
  const WorldModel & wm = agent->world() ; 
  const double dist_thr = 35.0 ; 
  int offender_unumm = 0 ;
  int blocker_unum = 0 ;
  is_blocker = false;
  
  if( !wm.audioMemory().mark().empty() && wm.audioMemory().mark().front().sender_ == wm.ourGoalieUnum() )
  {
    blocker_unum = wm.audioMemory().mark().front().offender_unums_[ 0 ];

    if( blocker_unum == wm.self().unum() )
    {
      is_blocker = true ;
      return is_blocker ; 

    }
    else if( wm.self().unum() < 9 )
    {
      is_blocker = false ;
      offender_unumm = wm.audioMemory().mark().front().offender_unums_[wm.self().unum()];
    }
    
    opp_target_unum = offender_unumm ;
    return true ;
  }
  
  return false ;
}
/*!*****************************************************************************************************************/
bool Bhv_Mark::say(PlayerAgent* agent , int blocker_unum)
{
  const WorldModel & wm = agent->world(); 
  int temp_off_unum[7];
  int j = 1;
  
  for( int i = 0 ; i < 7 ; i++ )
    temp_off_unum[i] = 0;
  
  temp_off_unum[0] = blocker_unum;
       
  for( int i = 2 ; i < 9 ; i ++  )
  {
      temp_off_unum[i-j] = offenders_unum[i];
  }
  
  if( agent->config().useCommunication() )
  {
    
    for( int i = 0 ; i < 7 ; i++ )
      last_marker_unum[i] = temp_off_unum[i];
    
    last_say_cycle = wm.time().cycle();
  
    return true;
  }
  else
    return false;
}
/*!*****************************************************************************************************************/
bool Bhv_Mark::get_best_edge(PlayerAgent* agent)
{
  return get_best_edge_mwbm(agent);
}

/*!*****************************************************************************************************************/

bool Bhv_Mark::get_best_edge_simple(PlayerAgent* agent, int & offender, double* score)
{
  int max_score = -32000;  
  
  std::vector<MarkEdge>::iterator it_end = cached_edges.end();
  
  for (std::vector<MarkEdge>::iterator it = cached_edges.begin() ; it != it_end ; it++)
  {
    
    if ( it->Marker()->unum() == agent->world().self().unum() && it->Score() > max_score)
    {
      
      max_score = it->Score();
      *score = max_score;
      offender = it->Offender()->unum();
    }  
  }
  
  if (max_score > -32000)
    return true;
  else 
    return false;
}

/*!*****************************************************************************************************************/

bool Bhv_Mark::get_best_edge_mwbm(PlayerAgent* agent)
{
  int opp_unum = -1;
  bool temp[MAXSIZE][MAXSIZE];
  int my_weight[MAXSIZE][MAXSIZE];
  int  weight[MAXSIZE][MAXSIZE];
  int unums[MAXSIZE][MAXSIZE][2];
  const WorldModel & wm = agent->world() ; 

    if( cached_edges.empty() )
      return false;


    std::vector<MarkEdge>::iterator it_end = cached_edges.end();
    for ( std::vector<MarkEdge>::iterator it = cached_edges.begin() ; it != it_end ; it++ )
    {    
      if( it->Offender()->unum() < 0 || it->Marker()->unum() < 0 )
	continue;
      
      my_weight[it->Offender()->unum()][it->Marker()->unum()] = it->Score();
      
      temp     [it->Offender()->unum()][it->Marker()->unum()] = true;
    }

     
      bool returning = false ; 
      getBipartiteGraph( 5 , 10 , my_weight );

      for(int iterator_marker = 0 ; iterator_marker < 10 ; iterator_marker ++)
      {

	if(match[iterator_marker] > -1)  
	  offenders_unum[ unums [ match [ iterator_marker ] ] [ iterator_marker ][ 0 ] ] 
			= unums [ match [ iterator_marker ] ] [ iterator_marker ][ 1 ];
	if(offenders_unum[ unums [ match [ iterator_marker ] ] [ iterator_marker ][ 0 ] ] )
	  returning = true ; 
      }
      return returning  ;
}


bool Bhv_Mark::dfs( int vertex_off , int max_count_off_mrk , int adj[MAXSIZE][MAXSIZE] , int mark[] , int s[] , int t[] )
{
  int i_marker;
  s[vertex_off] = 0;
  mark[vertex_off] = 0;
  
  for( i_marker= 0 ; i_marker < max_count_off_mrk ; i_marker++ )
    if( adj[vertex_off][i_marker] )
      if( t[i_marker] = 1 , match[i_marker] == 0 || ( mark[match[i_marker]] && !dfs( match[i_marker] , max_count_off_mrk , adj , mark , s , t ) ) )
	return match[i_marker] = vertex_off , true;
      
      return false;
    
}

bool Bhv_Mark::matching( int max_count_off_mrk , int adj[MAXSIZE][MAXSIZE] , int s[] , int t[] , int matching_checker[] )
{ 
  int i_vertex_off;
  int max = 0;
  int mark[MAXSIZE];

  for( i_vertex_off = 0 ; i_vertex_off < max_count_off_mrk ; i_vertex_off++ )
    if( matching_checker[i_vertex_off] && mark[i_vertex_off] && !dfs( i_vertex_off , max_count_off_mrk , adj , mark , s , t ) )
    {
      matching_checker[i_vertex_off] = 1;
    }
    
    for( i_vertex_off = 0 ; i_vertex_off < max_count_off_mrk ; i_vertex_off++ )
      if( !matching_checker[i_vertex_off] )
	return true;
      
      return false;
}

void Bhv_Mark::getBipartiteGraph( int count_offender , int count_marker , int weight[MAXSIZE][MAXSIZE] )
{
  
  int i, j;
  int size = 0;
  int matching_checker[MAXSIZE], s[MAXSIZE], t[MAXSIZE];
  int cover[2][MAXSIZE];
  int adj[MAXSIZE][MAXSIZE];
  for(int i1 = 0 ; i1 < MAXSIZE ; i1 ++)
  {
    matching_checker[i1] = 0 ;
    match[i1] = -1 ;
   for(int j1 = 0 ; j1 < MAXSIZE ; j1 ++ )
     adj[i1][j1] = 0 ;
  } 

  int max_count_off_mrk = 0 ;
  max_count_off_mrk = std::max (count_offender , count_marker );
  
  for (i = 1; i < max_count_off_mrk; i++)
  {
    int index = 0;
    for (j = 0; j < max_count_off_mrk; j++)
      if (weight[i][index] <= weight[i][j])	
	index = j;
	cover[0][i] = 0;
	cover[1][i] = weight[i][index];
	adj[i][index] = 1;
  }
  
  
  while (!matching(max_count_off_mrk, adj, s, t, matching_checker))
  {
    for(int i = 0 ; i < MAXSIZE ; i ++)
    {
      s[i] = 0 ;
      t[i] = 0 ;
    }
    matching(max_count_off_mrk, adj, s, t, matching_checker);
    int min = INF;
    
    for (i = 0; i < max_count_off_mrk; i++)
      for (j = 0; j < max_count_off_mrk; j++)
	if (s[i] && !t[j] && !adj[i][j])
	  if (cover[0][i] - cover[1][j] + weight[i][j] < min)
	    min = cover[0][i] - cover[1][j] + weight[i][j];
	  
	  for (i = 0; i < max_count_off_mrk; i++)
	    if (s[i])
	      cover[0][i] -= min;
	    
	    for (i = 0; i < max_count_off_mrk; i++)
	      if (t[i])
		cover[1][i] += min;
	      
	      for (i = 0; i < max_count_off_mrk; i++)
		for (j = 0; j < max_count_off_mrk; j++)
		  if ((!s[i] && !t[j]) || (s[i] && t[j]))
		    if ((cover[0][i] - cover[1][j]+weight[i][j])==1)
		      adj[i][j] = 1;
		    else
		      adj[i][j] = 1;
  } 
}



/*!*****************************************************************************************************************/

int Bhv_Mark::MarkEdge::area_danger(PlayerAgent * agent)
{
  const WorldModel & wm = agent->world(); 
  Vector2D offender_pos = Vector2D( 0 , 0 );
  Vector2D marker_pos   = Vector2D( 0 , 0 );
  int area_danger_value = 0 ; 
  if( offender->posValid() )
  {
    offender_pos = offender->pos() ; 
    marker_pos   =Strategy::i().getPosition( marker->unum() ) ;
   }
   else
     return 0 ; 
    const Strategy::BallArea offender_area = Strategy::get_ball_area( offender_pos );
    const Strategy::BallArea marker_area   = Strategy::get_ball_area( marker_pos   );
  switch( offender_area )
  {
    case Strategy::BA_CrossBlock:
      switch( marker_area )
      {
	case Strategy::BA_CrossBlock:
	  area_danger_value = 5 ;
	  break;
      
	case Strategy::BA_Danger:
	  area_danger_value =  5 ;
	  break;
      
	case Strategy::BA_DribbleBlock : 
	  area_danger_value = -4;
	  break;
      
	case Strategy::BA_DefMidField : 
	  area_danger_value =  -2;
	  break;
      }
    break ; 
    
    case Strategy::BA_Danger:
    switch( marker_area )
    {
      case Strategy::BA_CrossBlock:
	  area_danger_value =  8;
	  break;
      
      case Strategy::BA_Danger:
	  area_danger_value =  -15;
	  break;
      
      case Strategy::BA_DribbleBlock : 
	  area_danger_value =  4;
	  break;
      
      case Strategy::BA_DefMidField : 
	  area_danger_value =  -10;
	  break;
      } 
      break ;
      
    case Strategy::BA_DribbleBlock : 
    switch( marker_area )
    {
	case Strategy::BA_CrossBlock:
	  area_danger_value =  -5;
	  break;
      
	case Strategy::BA_Danger:
	  area_danger_value =  0;
	  break;
      
	case Strategy::BA_DribbleBlock : 
	  area_danger_value =  -15;
	  break;
      
	case Strategy::BA_DefMidField : 
	  area_danger_value =  7;
	  break;
      }
      break ;
      
    case Strategy::BA_DefMidField : 
    switch( marker_area )
    {
	case Strategy::BA_CrossBlock:
	  area_danger_value =  5;
	  break;
      
	case Strategy::BA_Danger:
	  area_danger_value =  -8;
	  break;
      
	case Strategy::BA_DribbleBlock : 
	  area_danger_value =  6;
	  break;
      
	case Strategy::BA_DefMidField : 
	  area_danger_value =  -15;
	  break;
     }
     break ;
 }
 return area_danger_value;
}
 

/*!*****************************************************************************************************************/
int Bhv_Mark::MarkEdge::home_pos_dist()
{
  Vector2D marker_home_pos = Strategy::i().getPosition( marker->unum() );
  return bound<int>( 0 , ceil (fabs( offender->pos().dist( marker_home_pos ) ) ), 25);
}

/*!*****************************************************************************************************************/

int Bhv_Mark::MarkEdge::offender_pos_count()
{
  return bound<int>(0 , offender->posCount() , 25 )  ;
}

/*!*****************************************************************************************************************/

int Bhv_Mark::MarkEdge::offender_to_marker_dist()
{
  const Vector2D our_goal_pos = ServerParam::i().ourTeamGoalPos();
  Vector2D marker_pos = marker->pos();
  Vector2D off_pos = offender->pos();
  Line2D line(our_goal_pos,off_pos);
  
  Vector2D team_pos_proj = line.projection(marker_pos);
  Vector2D opp_team = marker_pos-off_pos;
  Vector2D opp_goal = our_goal_pos - off_pos;
  double dist =( (team_pos_proj-off_pos).r() )*sign(opp_team.x*opp_goal.x+opp_team.y*opp_goal.y+0.1);
  double scale = 300.0/(area_danger(agent)+120.0);
  dist += sign(dist)*bound<double>(0.0,fabs((team_pos_proj-marker_pos).r())*scale,15.0);
  dist = bound<double>(-150,dist,150);
  double smart_dist = 0.0;
  //smart_dist = smart_dist;
  
  smart_dist = -fabs((marker_pos-off_pos).r());

  return bound<int>( 0 , floor (fabs (offender->pos().dist( marker->pos() ) ) )  , 25);
  return smart_dist;
  
}   

/*!*****************************************************************************************************************/
void Bhv_Mark::MarkEdge::sort_descending(std::vector< Vector2D > &vec_be_sorted, bool x_sort)
{
  if(x_sort){
    for(int i = 0 ; i < vec_be_sorted.size() ; i ++ )
      for(int j = 0 ; j < i ; j ++ ){
        Vector2D temp ; 
	if(vec_be_sorted.at(i).x < vec_be_sorted.at(j).x ){
	  temp = vec_be_sorted.at(j);
	  vec_be_sorted.at(j) = vec_be_sorted.at(i);
	  vec_be_sorted.at(i) = temp ; 
	}
      }
  }else{
    for(int i = 0 ; i < vec_be_sorted.size() ; i ++ )
      for(int j = 0 ; j < i ; j ++ ){
        Vector2D temp ; 
	if(vec_be_sorted.at(i).y < vec_be_sorted.at(j).y ){
	  temp = vec_be_sorted.at(j);
	  vec_be_sorted.at(j) = vec_be_sorted.at(i);
	  vec_be_sorted.at(i) = temp ; 
	}
      }
  }
    
} 

/*!*****************************************************************************************************************/
bool Bhv_Mark::block_decision(PlayerAgent* agent , int & teammate_blocker)
{

const WorldModel & wm = agent->world() ;
const PlayerPtrCont & teammates = wm.teammatesFromBall() ;
double score = 0  , max_score = 0 , score_threshold = 0; 
const Vector2D center_goal  ( -ServerParam::i().pitchHalfLength() , 0.0 ) ;

const PlayerObject * opponent_baller = wm.opponentsFromBall().front();
PlayerPtrCont::const_iterator teammate_it_end = teammates.end();
    for (PlayerPtrCont::const_iterator team_it = teammates.begin();
	 team_it != teammate_it_end;
	 team_it ++)
    {
      const Vector2D team_pos = (*team_it)->pos() ;

	double dist_to_opp = fabs( (*team_it)->pos().dist( opponent_baller->pos() ) );

	score = dist_to_opp * (-9.0)
	      + area_danger( agent , *team_it ) * 1.0;

	if(score > score_threshold && max_score < score && ( *team_it )->unum() != -1 )
	{
	  max_score = score ; 
	  teammate_blocker = ( *team_it )->unum() ; 
	}

      }
    last_blocker = teammate_blocker ; 
    
    return (max_score != -30000 ) ;
}
/*!*****************************************************************************************************************/
double Bhv_Mark::area_danger(PlayerAgent* agent , PlayerObject * marker)
{
  const WorldModel & wm = agent->world(); 
  PlayerObject *offender = wm.opponentsFromBall().front();
  Vector2D offender_pos = Vector2D( 0 , 0 );
  Vector2D marker_pos   = Vector2D( 0 , 0 );
  double area_danger_value = 0 ; 
  if( offender->posValid() && marker->posValid() )
  {
    offender_pos = offender->pos() ; 
    marker_pos   = Strategy::i().getPosition( marker->unum() ) ;
  }
  else
     return 0 ;
    const Strategy::BallArea offender_area = Strategy::get_ball_area( offender_pos );
    const Strategy::BallArea marker_area   = Strategy::get_ball_area( marker_pos   );
  switch( offender_area )
  {
    case Strategy::BA_CrossBlock:
      switch( marker_area )
      {
	case Strategy::BA_CrossBlock:
	  area_danger_value = 5 ;
	  break;
      
	case Strategy::BA_Danger:
	  area_danger_value =  -2 ;
	  break;
      
	case Strategy::BA_DribbleBlock : 
	  area_danger_value = 3;
	  break;
      
	case Strategy::BA_DefMidField : 
	  area_danger_value =  -2;
	  break;
	  
	default :
	  break;
      }
    break ; 
    
    case Strategy::BA_Danger:
    switch( marker_area )
    {
      case Strategy::BA_CrossBlock:
	  area_danger_value =  -8;
	  break;
      
      case Strategy::BA_Danger:
	  area_danger_value =  15;
	  break;
      
      case Strategy::BA_DribbleBlock : 
	  area_danger_value =  -4;
	  break;
      
      case Strategy::BA_DefMidField : 
	  area_danger_value =  10;
	  break;
	  
	default :
	  break;
    }
      break ;
      
    case Strategy::BA_DribbleBlock : 
    switch( marker_area )
    {
	case Strategy::BA_CrossBlock:
	  area_danger_value =  -5;
	  break;
      
	case Strategy::BA_Danger:
	  area_danger_value =  0;
	  break;
      
	case Strategy::BA_DribbleBlock : 
	  area_danger_value =  -15;
	  break;
      
	case Strategy::BA_DefMidField : 
	  area_danger_value =  7;
	  break;
	  
	default :
	  break;
      }
      break ;
      
    case Strategy::BA_DefMidField : 
    switch( marker_area )
    {
	case Strategy::BA_CrossBlock:
	  area_danger_value =  -5;
	  break;
      
	case Strategy::BA_Danger:
	  area_danger_value =  8;
	  break;
      
	case Strategy::BA_DribbleBlock : 
	  area_danger_value =  -6;
	  break;
      
	case Strategy::BA_DefMidField : 
	  area_danger_value =  15;
	  break;
	  
	default :
	  break;
     }
     break ;
	  
     default :
      break;
 }
 return area_danger_value;
}
