/*
    <one line to give the program's name and a brief idea of what it does.>
    Copyright (C) 2012  Mohammadhossein Malmir & Shahin Boluki :
                        <mhmalmir@gmail.com>   &  <sh.boluki@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_positioning.h"
#include "simmaldribble.h"
#include "withoutball.h"
#include "bhv_escape.h"
#include "bhv_mark.h"
#include "bhv_mark_action.h"
#include "bhv_block.h"

#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_ball_and_player.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/common/audio_memory.h>

#include "neck_offensive_intercept_neck.h"

#define DEBUG

using namespace rcsc ;

///*************************************************************************
bool Bhv_Positioning::execute( PlayerAgent * agent )
{
  
    dlog.addText( Logger::TEAM,__FILE__": Bhv_Positioning" );
    const WorldModel & wm = agent->world();
    int opp_target_unum = 0;
    Vector2D target = Vector2D::INVALIDATED ;
    double dash_power = ServerParam::i().maxDashPower();
    bool can = true ;
    double dist_thr = 1.5;
    
    PosType Type = action( agent , target , dist_thr , dash_power , opp_target_unum ); 

    ///Action
//     double xxxx = -60.0;
           switch (Type)
           {
                  case INTERCEPT:
                    if( !Body_Intercept().execute( agent ) )
                      can = false;
                    break;
                  case HOME_POS:
                  case MARK:
		  case BLOCK:
		    if ( !Body_GoToPoint( target , dist_thr, dash_power ).execute( agent ) )
		      Body_TurnToBall().execute( agent );
		    break;
		  case WITHOUT_BALL:
                    break;
           }
     ///View
           switch (Type)
           {
                  case INTERCEPT:
                  case HOME_POS:
		  case MARK:
                  case BLOCK:
		  case WITHOUT_BALL:
		   agent->setNeckAction( new Neck_TurnToBallOrScan() );
		   break;
           }
        
            return can;

}

///*************************************************************************

Bhv_Positioning::PosType Bhv_Positioning::action( PlayerAgent * agent ,
						  Vector2D & target ,
						  double & dist_thr ,
						  double & dash_power ,
						  int & view_number )
{
    const WorldModel & wm = agent->world();
    Vector2D homePos =  HomePos(agent) ;
    Vector2D markTarget =  HomePos(agent) ;
    Vector2D blockTarget =  HomePos(agent) ;
    Vector2D guardTarget =  HomePos(agent) ;
    Vector2D dangermarkTarget = HomePos(agent);
    double maxPower = ServerParam::i().maxDashPower();
    double markPower = ServerParam::i().maxDashPower();
    double blockPower = ServerParam::i().maxDashPower();
    double guardPower = ServerParam::i().maxDashPower();
    double dangermarkPower = ServerParam::i().maxDashPower();
    dash_power = maxPower ;
    dist_thr = 1.2 ;
    bool is_blocker = false ;
    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();
    bool should_not_intercept = (!wm.audioMemory().dribble().empty() &&
				  wm.audioMemory().dribble().front().sender_  != wm.self().unum() &&
				  wm.time().cycle() - wm.audioMemory().dribbleTime().cycle() <=
				  wm.audioMemory().dribble().front().queue_count_ );
    
 
     if ( ( ! wm.existKickableTeammate()
          && ( self_min <= 3 || ( self_min <= mate_min && self_min < opp_min + 3 ) )
          && ! should_not_intercept ) )
          return INTERCEPT ;
    
  bool MarkDec = Bhv_Mark().markExecute( agent , view_number , is_blocker );
  

     if(MarkDec && !wm.opponentsFromBall().empty() && 
          (
           ( is_blocker )
            || wm.opponentsFromBall().at(0)->unum() == view_number
          )
       )
          {
	    if( Bhv_Block().Blockexecute( agent , blockTarget ) )
	    {
	      dash_power = blockPower; 
	      target = blockTarget;
	      return BLOCK;
	    }
          }

   bool MarkAct = false;
    MarkAct = Bhv_Mark_Action( view_number ).markActionExecute( agent , markTarget , markPower );   

    
     if( ( (wm.self().unum()<6 ) || (wm.self().unum()>5) ) 
       && MarkDec && MarkAct &&
          wm.self().unum() > 1 && wm.self().unum() < 9 &&
         !wm.opponentsFromBall().empty() )
	  {
	    dash_power = markPower ;
	    target = markTarget ;
             return MARK ;
          }

    if ( WithoutBall().execute(agent) )
      return WITHOUT_BALL;

     target = homePos ;
      
     return HOME_POS ;
}
///*************************************************************************

Vector2D Bhv_Positioning::HomePos( PlayerAgent * agent )
{
  const WorldModel & wm = agent->world();
  return Strategy::i().getPosition( wm.self().unum() );
}
