/*!
    <SimMal Dribble Class.>
    Copyright (C) 2011  <Mohammadhossein Malmir> <mhmalmir@gmail.com> , <Mohammad Simchi> <m.simchi@yahoo.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define DEBUG

#include "simmaldribble.h"

#include "bhv_penalty_kick.h"

#include <rcsc/action/body_dribble.h>

#include <rcsc/action/body_kick_to_relative.h>

#include <rcsc/common/server_param.h>

#include <rcsc/geom/sector_2d.h>

#include <rcsc/player/penalty_kick_state.h>

using namespace rcsc;

bool     penalty_mode;
int 	 penalty_time;

//******************************************************************

bool SimMalDribble::dribbleExecute( PlayerAgent * agent ,
				    Vector2D SM_target_point ,
				    double SM_dist_thr ,
				    double SM_dash_power ,
				    int SM_dash_count ,
				    bool SM_dodge ,
				    bool SM_speedy , 
				    bool SM_Danger )
{
  const PlayerPtrCont & opps = agent->world().opponentsFromBall();
  const PlayerObject * nearest_opp = ( opps.empty()
				     ? static_cast< PlayerObject * >( 0 )
				     : opps.front() );
  const AngleDeg nearest_opp_dir = ( nearest_opp
				   ? nearest_opp->rpos().th()
				   : 1000.0 );
  
  AngleDeg angle_dribble = nearest_opp_dir + AngleDeg(180);
  
  if( SM_Danger )
  {
    return Body_KickToRelative( (SM_target_point - agent->world().self().pos()).r() , 
				angle_dribble - agent->world().self().body() , 
				true ).execute( agent );
  }
    
  return Body_Dribble(SM_target_point ,
				    SM_dist_thr ,
				    SM_dash_power ,
				    SM_dash_count ,
				    SM_dodge ,
				    SM_speedy ).execute( agent );
}

//******************************************************************

bool SimMalDribble::virtualFastDribbleExecute( PlayerAgent * agent ,
					       Vector2D & SM_target_point ,
					       double & SM_dist_thr ,
					       double & SM_dash_power ,
					       int & SM_dash_count ,
					       bool & SM_dodge ,
					       bool & SM_speedy, 
					       bool &SM_Danger  )
{
  const WorldModel & wm = agent->world(); 
  
  const Vector2D my_pos = wm.self().pos();
  
  if( wm.self().stamina() < 3300.0 )
    return false;
  
  int step = 12;
  
  Vector2D fast_dibble_point = dodgePoint( agent , step , 22.0 );
  if( isInLand( fast_dibble_point ) )
  {
    SM_target_point = fast_dibble_point;
    SM_dist_thr = 1.0;
    SM_dash_power = ServerParam::i().maxDashPower();
    SM_dash_count = 1;
    SM_dodge = false;
    SM_speedy = true;
    SM_Danger = false;
    return true;
  }
  
  return false;
}

//******************************************************************

bool SimMalDribble::virtualDribbleExecute( PlayerAgent * agent ,
					   bool pass,
				    Vector2D & SM_target_point ,
				    double & SM_dist_thr ,
				    double & SM_dash_power ,
				    int & SM_dash_count ,
				    bool & SM_dodge ,
				    bool & SM_speedy , 
		                    bool &SM_Danger )
{
  const WorldModel & wm = agent->world();
  
  if( wm.gameMode().isPenaltyKickMode() )
  {	
    Vector2D target_point = penalty_mode_dribble( agent );
    
    if( isInLand( target_point ) )
    {
      SM_target_point = target_point;
      SM_dist_thr = 2.0;
      SM_dash_power = ServerParam::i().maxDashPower();
      SM_dash_count = 6;
      SM_dodge = true;
      SM_speedy = true;
      SM_Danger = false;
      return true;
    }
    
    return false;
  }
  
  if( !wm.self().isKickable() )
    return false;
  
  const Vector2D my_pos = wm.self().pos();
  
  const PlayerPtrCont & opps_from_ball = wm.opponentsFromBall();
  
  const PlayerObject * nearest_opp_to_ball = ( opps_from_ball.empty()
					     ? static_cast< PlayerObject * >( 0 )
					     : opps_from_ball.front() );
  
  if( !nearest_opp_to_ball )
  {
    //std::cout<<wm.time().cycle()<<"\t:\toooo\n";
    return false;
  }
      
  if( wm.self().stamina() < 1000.0 )
    return false;
  
  const double nearest_opp_dist = ( nearest_opp_to_ball ) ? nearest_opp_to_ball->distFromSelf() : 1000.0;
  
  int step = 10;
  
  int my_side = ( my_pos.y >= 0.0 ) ? +1 : -1;
  
  Vector2D top_left = ( my_side == -1 ) ? my_pos : Vector2D( my_pos.x , -ServerParam::i().goalWidth() / 2 );
  Vector2D bottom_right = ( my_side == -1 ) ?
			  Vector2D( ServerParam::i().theirTeamGoalLineX() , ServerParam::i().goalWidth() / 2 ) : 
			  Vector2D( ServerParam::i().theirTeamGoalLineX() , my_pos.y ) ;
			  
  bool fast_dribble = false;
  
  if( my_pos.x >= 1.0 )
  {
    Rect2D rect( top_left , bottom_right );
    
    if( !agent->world().existOpponentIn<Rect2D>( rect , 6 , false ) )
    {
      fast_dribble = true;
    }
  }
  
  Vector2D fast_dibble_point = dodgePoint( agent , step , 10.0 );
  if( isInLand( fast_dibble_point ) )
  {
    SM_target_point = fast_dibble_point;
    SM_dist_thr = 1.0;
    SM_dash_power = ServerParam::i().maxDashPower();
    SM_dash_count = 1;
    SM_dodge = false;
    SM_speedy = true;
    SM_Danger = false;
    return true;
  }

    return Dribble( agent ,
		       SM_target_point ,
		       SM_dist_thr ,
		       SM_dash_power ,
		       SM_dash_count ,
		       SM_dodge ,
		       SM_speedy, 
		       SM_Danger  );

}

//******************************************************************

bool SimMalDribble::Dribble( PlayerAgent * agent ,
				    Vector2D & SM_target_point ,
				    double & SM_dist_thr ,
				    double & SM_dash_power ,
				    int & SM_dash_count ,
				    bool & SM_dodge ,
				    bool & SM_speedy,
				    bool & SM_Danger)
{
  const WorldModel & wm = agent->world();
  
  const PlayerPtrCont & opps = wm.opponentsFromBall();
  
  const PlayerObject * nearest_opp = ( opps.empty()
				     ? static_cast< PlayerObject * >( 0 )
				     : opps.front() );
	
  const double nearest_opp_dist = ( nearest_opp
                                  ? nearest_opp->distFromBall()
                                  : 1000.0 );
  
  const Vector2D nearest_opp_pos = ( nearest_opp
                                   ? nearest_opp->pos()
                                   : Vector2D( 1000.0, 0.0 ) );
  
  const AngleDeg nearest_opp_dir = ( nearest_opp
				   ? nearest_opp->rpos().th()
				   : 1000.0 );
  
  const Vector2D my_pos = wm.self().pos();
  
  const AngleDeg my_body_angle = wm.self().body();
  
  const int my_side = ( ( my_pos.y < 0.0 ) ? -1 : +1 );
 
  const AngleDeg base_angle = baseAngle( agent , my_pos );
  
  const Vector2D dirac_pos( ServerParam::i().pitchHalfLength() ,
		          ( ServerParam::i().DEFAULT_GOAL_WIDTH / 2.0 + 10.0 ) * my_side );
	
  Vector2D target_point = my_pos;
  AngleDeg dribble_angle = my_body_angle;
  
  AngleDeg front_dribble_angle = my_body_angle;
  
  const double R = bound<double>( 5.0 , nearest_opp_dist * 2.0 , 8.0 );
  
  Sector2D front_sector ( my_pos , 0.1 , R , base_angle - 90.0 , base_angle + 90.0 );
	  
  int front_opps_num  = wm.countOpponentsIn<Sector2D>( front_sector  , 5 , true );
  
  int i , j , k;
	  
  std::vector<AngleDeg> front_angles;
  
  AngleDeg front_max_angle = AngleDeg(-10000.0);
	  
  if ( front_opps_num > 0 )
  {
    PlayerPtrCont::const_iterator opps_end = opps.end();
    for ( PlayerPtrCont::const_iterator opp = opps.begin();
          opp != opps_end;
	  ++opp )
    {
      if( (*opp)->posCount() > 10 )
	continue;
	  
      if( (*opp)->isGhost() && (*opp)->posCount() >= 4 )
	continue;
	  
      if( front_sector.contains( (*opp)->pos() ) )
	  front_angles.push_back( (*opp)->rpos().th() );
    }
			   
    front_angles.push_back( base_angle + AngleDeg(+90.0) );
    front_angles.push_back( base_angle + AngleDeg(-90.0) ); 
    int front_angles_len = front_angles.size();
	  
    AngleDeg a = my_body_angle;
	  
    for( j = front_angles_len - 1 ; j > 0 ; j-- )
      for( k = 0 ; k < j ; k++ )  
        if( front_angles.at(k).degree() > front_angles.at(k+1).degree() )
        {
	  a = front_angles.at(k);
	  front_angles.at(k) = front_angles.at(k+1);
	  front_angles.at(k+1) = a;
        }

				
    std::vector<AngleDeg> front_dir_diff;
	  
    for( i = 0 ; i < front_angles_len - 1 ; i++ )
     front_dir_diff.push_back( ( front_angles.at(i) - front_angles.at(i+1) ).abs() );
    
    int front_max = 0;
	  
    for( i = 0 ; i < front_angles_len -1 ; i++ )
      if ( front_dir_diff.at(i).degree() > front_max_angle.degree() )
      {
        front_max_angle = front_dir_diff.at(i);
        front_max = i;
      }

    if( front_max_angle.degree() > 120 )
    {
      front_dribble_angle = ( ( ( front_angles.at(front_max) + AngleDeg(62) ) - base_angle ).abs() <
			    ( ( front_angles.at(front_max+1) - AngleDeg(62) ) - base_angle ).abs() ) ?
			      ( front_angles.at(front_max) + AngleDeg(62) ) :
			      ( front_angles.at(front_max+1) - AngleDeg(62) );
    }
	  
    else
      front_dribble_angle = ( front_angles.at(front_max).degree() + front_angles.at(front_max+1).degree() ) / 2;
  }
  
  Vector2D front_final_pos =  Vector2D::polar2vector( 1.5 , front_dribble_angle );
  
  if( isInLand( target_point ) )
  {
    SM_target_point = target_point;
    SM_dist_thr = 0.75;
    SM_dash_power = ServerParam::i().maxDashPower();
    SM_dash_count = 2;
    SM_dodge = true;
    SM_speedy = false;
    SM_Danger = false;
    return true;
  }
  
  return false;
}

//******************************************************************

int SimMalDribble::cyclesObjectToReachPos( PlayerAgent * agent , PlayerObject * o , Vector2D pos , bool is_self )
{
  if( is_self )
  {
    const WorldModel & wm = agent->world();
    Vector2D my_vel = wm.self().vel();
    Vector2D my_pos = wm.self().pos();
    double my_dist = my_pos.dist( pos );
    
    return (ceil( my_dist / my_vel.r() ) );
  }
  
  else
  {
    double object_vel = (*o).vel().r();
    
    Vector2D object_pos = (*o).pos();
    double dist = object_pos.dist( pos );		
    return ( ceil( dist / object_vel ) ); 
  }
}

//******************************************************************

Vector2D SimMalDribble::dodgePoint( PlayerAgent * agent , int step , double length )
{
  const WorldModel & wm = agent->world(); 
  
  const PlayerPtrCont & opps = wm.opponentsFromBall();
  
  const Vector2D my_pos = wm.self().pos();
  
  Vector2D base_pos = my_pos;
 
  AngleDeg base_angle = baseAngle( agent , my_pos );

  Vector2D dodge_point = my_pos;
  
  Vector2D rel_temp = my_pos;
  Vector2D temp = my_pos;
  
  double score = 0.0;
  double best_score = 10000.0;

  const double dir_diff_to_base_angle_weight = -1.0;
  const double dir_diff_to_near_opp_weight = -1.0;
  const double dist_diff_to_goal_weight = -1.0;
  const double dist_diff_to_near_opp_weight = -1.0;
  
  int i , j , l , m;
  unsigned int k;
  double n;
  
  int goalie_row;
  
  std::vector<double> cycle_dist;
  
  PlayerPtrCont::const_iterator opp = opps.begin();
      
  //PlayerObject * o = (PlayerObject*)(*opp);
      
  bool temper = false;
      
  const PlayerPtrCont::const_iterator opps_end = opps.end();
      
  int opponent_reach_cycle;
  int me_reach_cycle;
  
  double r = length;

  int min = 0;
  double min_cycle = 10000.0;
  
  for( i = 8 ; i <= r ; i++ )
  {    
    base_angle = baseAngle( agent , my_pos );
    base_pos = my_pos;
    
    for( l = 1 ; l >= -1 ; l -= 2 )
      for( j = base_angle.degree() , m = 0 ; m < step ; j += ( 3 * l ) , m++ )
      {
        cycle_dist.clear();
      
	goalie_row = -1;
	
        rel_temp.setPolar( i , j );
        temp = rel_temp + my_pos;
      
        if( !isInLand( temp ) )
        {
	  //std::cout<<wm.time().cycle()<<"\t"<<temp<<"\tis not in land\n";
	  continue;
        }
	
        if( wm.existOpponentIn<Sector2D>( Sector2D( wm.ball().pos() , 0.6 , i , j - 1 , j + 1 ) , 4 , true ) )
        {
   	  //std::cout<<wm.time().cycle()<<"\t"<<temp<<"\texist opp\n";
 	  continue;
        }
      
        temper = false;
	
	me_reach_cycle = cyclesObjectToReachPos( agent , *opp , temp , true );
      
        for ( opp = opps.begin() ; opp != opps_end ; ++opp )
        {
	  if( (*opp)->posCount() > 10 )
	    continue;
	
	    opponent_reach_cycle = cyclesObjectToReachPos( agent , *opp , temp , false );

	  
	  if( opponent_reach_cycle <= me_reach_cycle )
	  {
	    //std::cout<<wm.time().cycle()<<"\t"<<temp<<"\topp unum\t"<<(*opp)->unum()<<"\tdiff reach cycle\t"<<me_reach_cycle - opponent_reach_cycle<<"\tBAD\n";
	    temper = true;
	    break;
	  }
	
	  cycle_dist.push_back( opponent_reach_cycle );
        }
      
        if(temper)
	  continue;
      
        min = -1;
        min_cycle = 10000.0;
      
        for( k = 0 ; k < cycle_dist.size() ; k++ )
	  if( cycle_dist.at(k) < min_cycle )
	  {
	    min = k;
	    min_cycle = cycle_dist.at(k);
	  }

	  score = ( rel_temp.th() - base_angle ).abs() * dir_diff_to_base_angle_weight + 
		  ( rel_temp.th() - opps.at( min )->rpos().th() ).abs() * dir_diff_to_near_opp_weight +
		    ServerParam::i().theirTeamGoalPos().dist( temp ) * dist_diff_to_goal_weight +
		    rel_temp.dist( opps.at( min )->rpos() ) * dist_diff_to_near_opp_weight;

        if( score < best_score )
        {
	  best_score = score;
	  dodge_point = temp;
        }
      }
  }  
    
  if( dodge_point != my_pos )
    return dodge_point;
  
  return Vector2D::INVALIDATED;
}


//******************************************************************

AngleDeg SimMalDribble::baseAngle( PlayerAgent * agent , Vector2D mypos )
{ 
  return 0.0;
}

//******************************************************************

bool SimMalDribble::isInLand( Vector2D pos )
{
  if( !pos.isValid() )
    return false;
  
  if( pos.absX() > ( ServerParam::i().pitchHalfLength() - 0.25 ) ||
      pos.absY() > ( ServerParam::i().pitchHalfWidth()  - 0.25 ) )
    return false;
  
  return true;
}
//******************************************************************

Vector2D SimMalDribble::penalty_mode_dribble ( rcsc::PlayerAgent * agent )
{
  const WorldModel & wm = agent->world(); 
  
  int goalie_unum = wm.theirGoalieUnum();
  
  if( goalie_unum < 1 || goalie_unum > 11 )
    return Vector2D::INVALIDATED;
  
  PlayerObject * opp_goalie = ( PlayerObject * ) wm.theirPlayer( goalie_unum );

  if( !opp_goalie || opp_goalie->ghostCount() > 6 )
    return Vector2D::INVALIDATED;
  
  const Vector2D my_pos = wm.self().pos();
  const int my_side = ( my_pos.y >= 0 ) ? +1 : -1;
  
  Vector2D base_pos = my_pos;  

  AngleDeg base_angle = baseAngle( agent , my_pos );
  
  Vector2D dodge_point = my_pos;
  
  Vector2D rel_temp = my_pos;
  Vector2D temp = my_pos;
  
  double score = 0.0;
  double best_score = 10000.0;
  
  const double dir_diff_to_base_angle_weight = -1.0;
  const double dir_diff_to_near_opp_weight = -1.0;
  const double dist_diff_to_goal_weight = -1.0;
  const double dist_diff_to_near_opp_weight = -1.0;
  
  int i , j , l , m;
  unsigned int k;
  double n;
  
  int goalie_row;
  
  std::vector<double> cycle_dist;
     
  bool temper = false;
      
  int opp_goaly_reach_cycle;
  int me_reach_cycle;
  
  double r = 7.0;
  
  int step = fabs( ServerParam::i().theirTeamGoalLineX() - my_pos.x ) * 0.3 ;
	
  int min = 0;
  double min_cycle = 10000.0;
  
  for( i = 2 ;  i <= r && opp_goalie->pos().dist( my_pos ) >= 5.0 ; i++ )
  {
    base_angle = baseAngle( agent , my_pos );
    base_pos = my_pos;
    
    for( l = 1 ; l >= -1 ; l -= 2 )
      for( j = base_angle.degree() , m = 0 ; m < step ; j += ( 3 * l ) , m++ )
      {      
        rel_temp.setPolar( i , j );
        temp = rel_temp + my_pos;
      
        if( !isInLand( temp ) )
        {
		  continue;
        }
	
        if( wm.existOpponentIn<Sector2D>( Sector2D( wm.ball().pos() , 0.6 , i , j - 1 , j + 1 ) , 4 , true ) )
        {
		  continue;
        }
	
		me_reach_cycle = cyclesObjectToReachPos( agent , opp_goalie , temp , true );
		
	    opp_goaly_reach_cycle = cyclesObjectToReachPos( agent , opp_goalie , temp , false );
	  
		if( opp_goaly_reach_cycle <= me_reach_cycle )
		{
		  continue;
		}

		score = ( rel_temp.th() - base_angle ).abs() * dir_diff_to_base_angle_weight + 
				( rel_temp.th() - opp_goalie->rpos().th() ).abs() * dir_diff_to_near_opp_weight +
				  ServerParam::i().theirTeamGoalPos().dist( temp ) * dist_diff_to_goal_weight +
				  rel_temp.dist( opp_goalie->rpos() ) * dist_diff_to_near_opp_weight ;


        if( score < best_score )
        {
		  best_score = score;
		  dodge_point = temp;
        }
      }
	} 
    
    
//   if( dodge_point != my_pos )
    return dodge_point;

  
}

